#############################################################################
## 
## GetGroebnerSpecial( polys, ORDER )
## 
## Computes parts of a Groebner basis.
## Currently unused.
## 
GetGroebnerSpecial := function( polys, ORDER )
    local i, subs, rems, poly;

    poly := Set(Filtered(polys, x -> x <> 0*x));

    for i in [2..6] do
        subs := Filtered(poly, x -> PolyLen(x) <= i );
        rems := Filtered(poly, x -> PolyLen(x) > i );
        subs := GetGroebner( subs, ORDER, 7 ); #!#!
        rems := List( rems, x -> PRR( x, subs, ORDER ) );
        poly := Concatenation(subs, rems );
        poly := Set(Filtered(poly, x -> x <> 0*x));
        Print("   special groeber : gb of length ",Length(subs)," - ");
        Print(Length(rems)," remaining \n");
    od;

    return polys;
end;

#############################################################################
## 
## HallInherit( n, gb )
## 
## Inherit the consistency relations from HallRec[n-1].cons 
## 
HallInherit := function( n, gb )
    local cons1, cons2, cons3, U, V, W, i, j, k;

    if not IsBound(HallRec[n-1].cons) then return fail; fi;

    HallRec[n].inht := [];

    # step 1: use U 
    cons1 := HallRec[n-1].cons;
    
    # step 2: use V
    U := Flat( T{[1..n]}{[1..n]}{[1..n]} );
    V := Flat( T{[2..n+1]}{[2..n+1]}{[2..n+1]} );
    cons2 := List( cons1, f -> Value( f, U, V ) );
    
    # step 3: use W
    W := 0*T{[1..n]}{[1..n]}{[1..n]};
    W[1][1][1] := T[1][1][1];
    for i in [3..n+1] do
        for j in [3..n+1] do
            for k in [3..n+1] do
                W[i-1][j-1][k-1] := T[i][j][k];
            od;
        od;
    od;
    for i in [3..n+1] do
        for j in [3..n+1] do
            W[1][i-1][j-1] := T[1][i][j];
            W[i-1][1][j-1] := T[i][1][j];
            W[i-1][j-1][1] := T[i][j][1];
        od;
    od;
    for i in [3..n+1] do
        W[1][1][i-1] := T[1][1][i];
        W[1][i-1][1] := T[1][i][1];
        W[i-1][1][1] := T[i][1][1];
    od;
    cons3 := List( cons1, f -> Value( f, U, Flat( W ) ) );

    HallRec[n].inht := Set( Concatenation( cons1, cons2, cons3 ) );

# Special version of Groebner bases for n >= 7
#    if gb and Length(HallRec[n].inht) > 0 and n < 7 then
#        Print(" get Groebner in HallInherit \n");
#        HallRec[n].inht := GetGroebner( HallRec[n].inht, ORDER );
#        Print(" ... done \n");
#    elif gb and n >= 7 then
#        Print(" get SPECIAL Groebner in HallInherit \n");
#        HallRec[n].inht := GetGroebnerSpecial( HallRec[n].inht, ORDER );
#        Print(" ... done \n");
#    fi;

    # compute Groebner basis if desired
    if gb and Length(HallRec[n].inht) > 0 then
        Print(" get Groebner in HallInherit \n");
        HallRec[n].inht := GetGroebner( HallRec[n].inht, ORDER, 7 ); #!#!
        Print(" ... done \n");
    fi;
    
    return true;
end;


#############################################################################
## 
## HallConsistency( n, gb )
## 
## Compute consistency relations from HallRec[n].mult and HallRec[n].inht.
## 
HallConsistency := function( n, gb )
    local A, B, C, F, R, G, L, T, i;

    if not IsBound( HallRec[n].mult ) then return fail; fi;

    # initialize variables
    A := List( [1..n], x -> Indeterminate( Rationals,
                                Concatenation( "a", String( x ) ) ) );
    B := List( [1..n], x -> Indeterminate( Rationals,
                                Concatenation( "b", String( x ) ) ) );
    C := List( [1..n], x -> Indeterminate( Rationals,
                                Concatenation( "c", String( x ) ) ) );

    # read multiplication polynomial F := m( A, B )
    F := List( [1..n], x -> A[x]+B[x]+HallRec[x].mult );

    # compute F( F( A, B ), C )
    R := List( F, f -> Value( f, Concatenation( A, B ),
                                Concatenation( F, C ) ) );

    # compute F( A, F( B, C ) )
    G := List( F, f -> Value( f, Concatenation( A, B ),
                                Concatenation( B, C ) ) );
    L := List( F, f -> Value( f, Concatenation( A, B ),
                                Concatenation( A, G ) ) );

    # read off consistency relations
    T := Filtered( R-L, x -> x <> 0*x );

    # read off coefficients
    for i in [1..n] do
        T := Flat( List( T, x ->
                        PolynomialCoefficientsOfPolynomial( x, A[i] ) ) );
        T := Flat( List( T, x ->
                        PolynomialCoefficientsOfPolynomial( x, B[i] ) ) );
        T := Flat( List( T, x ->
                        PolynomialCoefficientsOfPolynomial( x, C[i] ) ) );
        T := Set(Filtered( T, x -> x <> 0*x ));
    od;

    # reduce with known consistency
    if Length(HallRec[n].inht) > 0 then 
        for i in [1..Length( T )] do
            T[i] := PRR( T[i], HallRec[n].inht, ORDER );
        od;
        T := Set(Filtered( T, x -> x <> 0*x ));
        T := Concatenation( HallRec[n].inht, Set( T ) );
    fi;

    HallRec[n].cons := T;

#    if gb and Length(HallRec[n].cons) > 0 and n < 7 then 
    if gb and Length(HallRec[n].cons) > 0 then 
        Print(" get Groebner in HallConsistency \n");
        HallRec[n].cons := GetGroebner( HallRec[n].cons, ORDER, 7 ); #!#!
        Print("Length of cons: ", Length(HallRec[n].cons), "\n");
        Print(" ... done \n");
#    elif gb and Length(HallRec[n].cons) > 0 and n >= 7 then 
#        Print(" get SPECIAL Groebner in HallConsistency \n");
#        HallRec[n].cons := GetGroebnerSpecial( HallRec[n].cons, ORDER );
#        Print(" ... done \n");
    fi;
    return true;
end;

#############################################################################
## 
## SolveRec( g, v )
## 
## Return the solution of the recurrence relation f( 0 ) = 0, 
## f( v+1 ) = f( v ) + g( v ), where g( v ) is a polynomial.
## 
## Input: v = indeterminate, g = coefficients of g( v )
## 
SolveRec := function( g, v )
    local f, j, w;
 
    f := 0*v;
    if Length( g ) = 0 then return f; fi;

    f := g[1]*v;
    if Length( g ) = 1 then return f; fi;

    for j in [1..Length( g )-1] do
        w := List( [0..j], l -> Bernoulli( j-l ) *
                                Binomial( j, l )*v^( l+1 )/( l+1 ) );
        f := f + g[j+1]*Sum( w );
    od;

    return f;
end;


#############################################################################
## 
## CollectSymbolic( n, R )
## 
CollectSymbolic := function( n, R )
    local v, stack, c, i, j, k, m;
    
    # set up
    R := R*x^0;
    
    # initialize
    v := StructuralCopy( A );
    v[n] := v[n] + B[n];
    stack := Reversed( List( [1..n-1], x -> [x, B[x]] ) );
    
    # collection
    while Length( stack ) > 0 do
        c := stack[Length( stack )];
        Unbind( stack[Length( stack )] );
        if c[2] <> 0*c[2] then
            i := c[1];
            v[i] := v[i] + c[2];
            for j in Reversed( [i+1..n-1] ) do
                for k in Reversed( [1..Length( R[i][j] )] ) do
                    if R[i][j][k] <> 0*R[i][j][k] then 
                        m := Value( R[i][j][k], [x, y], [v[j], c[2]] );
                        if k = n then 
                            v[n] := v[n] + m;
                        elif m <> 0*m then 
                            Add( stack, [k, m] );
                        fi;
                    fi;
                od;
                v[j] := 0;
            od;
        fi;
    od;
    
    return v;
end;


#############################################################################
## 
## DetermineHallMult( n )
## 
## Determine HallRec[n].mult from HallRec[n-1].
## 
DetermineHallMult := function( n )
    local R, i, j, S, W, V, k, a, b, s, l, g, h, v;

    if not IsBound(HallRec[n-1]) then return fail; fi;

    # initialize conjugacy polynomials R[i][j][k]
    R := NullMat( n, n );
    for i in [1..n] do
        for j in [1..n] do
            R[i][j] := 0*[1..n];
        od;
    od;

    # known conjugacy polynomials in <g_2, ..., g_n>
    S := NullMat( n-1, n-1 );
    for i in [2..n] do
        for j in [2..n] do
            S[i-1][j-1] := T[i][j]{[2..n]};
        od;
    od;

    # known conjugacy polynomials in <g_1, g_3, ..., g_n>
    W := 0*S;
    for i in [3..n] do
        for j in [3..n] do
            for k in [3..n] do
                W[i-1][j-1][k-1] := T[i][j][k];
            od;
        od;
    od;
    for i in [3..n] do
        for j in [3..n] do
            W[1][i-1][j-1] := T[1][i][j];
            W[i-1][1][j-1] := T[i][1][j];
            W[i-1][j-1][1] := T[i][j][1];
        od;
    od; 
    for i in [3..n] do
        W[1][1][i-1] := T[1][1][i];
	W[1][i-1][1] := T[1][i][1];
        W[i-1][1][1] := T[i][1][1];
    od;
    W[1][1][1] := T[1][1][1];

    # known conjugacy polynomials in <g_1, .., g_n>/<g_n>
    V := T{[1..n-1]}{[1..n-1]}{[1..n-1]};
    
    # compute conjugacy polynomials
    for i in [1..n] do
        for j in [i+1..n] do
            Print( "    compute r[i][j] for ", i, " by ", j, "\n" );
            if i <> 1 then 
                # computation in <g_2, ..., g_n>
                a := 0*[2..n]; a[j-1] := x;
                b := 0*[2..n]; b[i-1] := y;
                R[i][j] := HallMult( S, -b, a );
                R[i][j] := HallMult( S, R[i][j], b );
                R[i][j] := Concatenation( [0], R[i][j] );
            elif i=1 and j > 2 then
                # computation in <g_1, g_3, ..., g_n>
                a := 0*[2..n]; a[j-1] := x;
                b := 0*[2..n]; b[1] := y;
                R[i][j] := HallMult( W, -b, a );
                R[i][j] := HallMult( W, R[i][j], b );
                R[i][j] := Concatenation( [0], R[i][j] );
            elif i=1 and j=2 then 

                # step one : x = 1 and k < n
                a := 0*[1..n-1]; a[2] := 1;
                b := 0*[1..n-1]; b[1] := y;
                R[1][2]{[1..n-1]} := HallMult( V, HallMult( V, -b, a ), b );

                # step two: x = 1 and k = n -- setup recursion
                s := List( [3..n], i -> Concatenation( "s", String( i ) ) ); 
                s := List( s, i -> Indeterminate( Rationals, i ) );
                s := Concatenation( [0], s );
                b := 0*[2..n]; b[1] := 1;
                l := HallMult( W, HallMult( W, -b, s ), b );

                a := Concatenation( [0], T[1][2]{[3..n]} );
                l := HallMult( W, a, l )[n-1];
                l := l - s[n-1];

                # step three: x = 1 and k = n -- solve recursion
                l := Value( l, s{[2..n-2]}, R[1][2]{[3..n-1]} );
                g := [];
                while l <> 0*l do
                    h := Value( l, [y], [0] );
                    Add( g, h );
                    l := ( l-h )/y;
                od;
                R[1][2][n] := SolveRec( g, y );

                # aribtrary x
                l := HallPower( S, R[1][2]{[2..n]}, x );
                R[1][2] := Concatenation( [0], l );

            fi;
        od;
    od;
    
    # compute multiplication polynomial via symbolic collection
    Print( "    starting symbolic collection \n" );
    HallRec[n].mult := CollectSymbolic( n, R )[n] - A[n] - B[n];
end;

#############################################################################
## 
## DetermineHallPower( n )
## 
## Determine HallRec[n].powr from HallRec[n-1] and HallRec[n].mult.
## 
DetermineHallPower := function( n )
    local k, i, j, w, g, h, v;

    if not IsBound(HallRec[n].mult) then return fail; fi;

    # set up
    k := 0*A{[1..n]};
    for i in [1..n-1] do
        k[i] := x*A[i] + HallRec[i].powr;
    od;
    w := HallMult( T, A{[1..n]}, k )[n] - A[n];
    
    # set up recurrence relation
    Print( "    extract terms  \n" );
    g := [];
    while w <> 0*w do
        h := Value( w, [x], [0] );
        Add( g, h );
        w := ( w-h )/x;
    od;

    # solve recurrence relation
    Print( "    solve recursion  \n" );
    k[n] := x*A[n] + SolveRec( g, x );
       
    HallRec[n].powr := k[n] - x*A[n];
end;


#############################################################################
## 
## HallReduce( n, flag )
## 
## Reduce mult or power polynomial either with consistency ( if known ) or
## with inherited consistency ( if known )
## 
HallReduce := function( m, flag )

    if IsBound( HallRec[m].cons ) and flag = "mult" then 
        HallRec[m].mult := PRR( HallRec[m].mult, HallRec[m].cons, ORDER );
    elif IsBound( HallRec[m].inht ) and flag = "mult" then 
        HallRec[m].mult := PRR( HallRec[m].mult, HallRec[m].inht, ORDER );
    fi;

    if IsBound( HallRec[m].cons ) and flag = "powr" then 
        HallRec[m].powr := PRR( HallRec[m].powr, HallRec[m].cons, ORDER );
    elif IsBound( HallRec[m].inht ) and flag = "powr" then 
        HallRec[m].powr := PRR( HallRec[m].powr, HallRec[m].inht, ORDER );
    fi;

end;


#############################################################################
## 
## HallConstruction( n )
## 
## Compute HallRec[n] from HallRec[n-1]
## 
HallConstruction := function( n, gb )
    
    if not IsBound( HallRec[n-1] ) then return fail; fi;
    
    # inherit consistency
    HallInherit( n, gb );
    
    # compute new multiplication polynomial
    Print( " - Step ", n, ": compute mult-polynomial \n" );
    DetermineHallMult( n );
    Print( "    got poly with ", PolyLen( HallRec[n].mult ), " summands\n" );
    
    # reduce mult polynomial 
    Print( " - Step ", n, ": reduce mult-polynomial via inherit \n" );
    HallReduce( n, "mult" );
    Print( "    mult poly has ", PolyLen( HallRec[n].mult ), " summands\n" );
    
    # compute new consistency relations
    Print( " - Step ", n, ": compute new consistency relations \n" );
    HallConsistency( n, gb );
        
    # reduce multiplication polynomial with new consistency relations
    Print( " - Step ", n, ": reduce mult polynomial via consistency\n" );
    HallReduce( n, "mult" );
    Print( "    mult poly has ", PolyLen( HallRec[n].mult ), " summands\n" );
    
    # compute new power polynomial
    Print( " - Step ", n, ": compute power polynomial \n" );
    DetermineHallPower( n );
    
    # reduce power polynomial with consistency relations
    Print( " - Step ", n, ": reduce power polynomial \n" );
    HallReduce( n, "powr" );
    Print( "    power poly has ", PolyLen( HallRec[n].powr )," summands\n" );
    
    # save computed polynomials if neccessary
    if not CheckFileExists( n ) then 
        Print( " - Step ", n, ": save computed polynomials \n" );
        SaveHallPolynomials( n );
    fi;
    
    Print( " - Step ", n, ": done \n\n" );
end;

