#############################################################################
## 
## AddName( var, string )
## 
## Add the name of a variable.
## 
## Input: var = considered variable; string = name of the variable
## 
AddName := function( var, string )
    NAMES[ExtRepPolynomialRatFun( var )[1][1]] := string;
end;


#############################################################################
## 
## SetupVariables( n )
## 
SetupVariables := function( n )
    local i, j, k;

    # reset NAMES
    NAMES := [];

    # initialize variables A[i] and B[i]
    A := List( [1..n], x -> 0 );
    B := List( [1..n], x -> 0 );
    for i in [1..n] do
        A[i] := Indeterminate( Rationals,
                                Concatenation( "a", String( i ) ) );
        AddName( A[i], Concatenation( "A[", String( i ), "]" ) );
        B[i] := Indeterminate( Rationals,
                                Concatenation( "b", String( i ) ) );
        AddName( B[i], Concatenation( "B[", String( i ), "]" ) );
    od;

    # initialize variables x and y
    AddName( x, "x" );
    AddName( y, "y" );

    # initialize HallRec
    # initialize variables T[i][j][k]
    T := NullMat( n, n );
    for i in [1..n] do
        T[i] := [];
        for j in [1..n] do
            T[i][j] := [];
            for k in [1..n] do
                T[i][j][k] := Concatenation( "t", 
                                String( i ), 
                                String( j ), 
                                String( k ) );
                T[i][j][k] := Indeterminate( Rationals, T[i][j][k] );
                AddName( T[i][j][k], Concatenation( "T[", 
                                String( i ), "][", 
                                String( j ), "][", 
                                String( k ), "]" ) );
            od;
        od;
    od;

    HallRec := List( [1..n], x -> rec( mult := 0, powr := 0, cons := [] ) );

end;


