#############################################################################
## 
## Runtime test: evaluation of Hall polynomials vs. Collection.
## Use subgroups of UT_6(Z) of Hirsch-length <= 7.
## 
## ngrp  = number of considered groups
## nprod = number of considered products
## 
RuntimeTestHallColl := function(ngrp, nprod)
    local tH, tC, G, U, h, t, W, w, a, b, p, q, t1, t2, i,j,k;
    
    tH := 0;
    tC := 0;

    G := UnitriangularPcpGroup(6, 0);
    
    # ngrp arbitrary presentations
    for i in [1..ngrp] do

        Print(" Looking for subgroup ", i, "...\n");
        while true do
            U := GroupByGenerators([Random(G),Random(G)]);
            h := HirschLength(U);
            if h <= 7 then 
                Print(" Found a subgroup of Hirsch length ", h, ".\n");
                break;
            fi;
        od;

        t := VectorByGroupCS(U);
        W := GroupByVector(h,t);
        w := Igs(W);
        
        # nprod arbitrary porducts with "large" coefficients
        for j in [1..nprod] do
            a := List([1..h], x -> Random([-1000..1000]));
            b := List([1..h], x -> Random([-1000..1000]));
            
            # answer by using Hall-polynomials
            t1 := Runtime();
            p := HallMult( t, a, b ); 
            t1 := Runtime() - t1;
            tH := tH + t1;
            
    	    a := MappedVector( a, w );
    	    b := MappedVector( b, w );

    	    # answer by using Collection
            t2 := Runtime();
            q := a*b; 
            t2 := Runtime() - t2;
            tC := tC + t2;
 
            Print("  Hall polynomials: ",t1,"ms, Collection: ",t2,"ms\n");
            
        od;
        Print("\n");
    od;
    
    Print(" Overall runtime: Hall polynomials: ", StringTime(tH),
                ", Collection: ", StringTime(tC), "\n");
    
    return [tH,tC];
end;

