#############################################################################
##  
##  init.g                hallpoly package                       Bettina Eick
##                                                             Alexander Cant
##  

#############################################################################
##
## Read singular package
##
sg := LoadPackage("singular");
if sg then
    # Method for computing a Groebner basis using degBound in Singular
    InstallOtherMethod( GroebnerBasis,
            "for an ideal in a poly ring", true,
            [ IsPolynomialRingIdeal, IsInt ], 0, 
            function ( I, degb )
                local input, out;

                Info( InfoSingular, 2, "running GroebnerBasis..." );

                # preparing the input for Singular
                input := Concatenation("degBound = ", String(degb), ";");
                Append( input, "ideal GAP_groebner = groebner( " );
                Append( input, ParseGapIdealToSingIdeal( I ) );
                Append( input, " );\n" );

                out := SingularCommand( input, "string (GAP_groebner)" );

                Info( InfoSingular, 2, "done GroebnerBasis." );

                return List( SplitString( out, ',' ), ParseSingPolyToGapPoly );
            end );
    SINGULARGBASIS := rec(
        name := "singular interface for GroebnerBasis",
        GroebnerBasis := function ( pols, O, degb )
            local  ipr, mcf, R, I;

            if IsPolynomialRingIdeal( pols )  then
                R := LeftActingRingOfIdeal( pols );
                pols := GeneratorsOfTwoSidedIdeal( pols );
            else
                R := DefaultRing( pols );
            fi;

            if IsMonomialOrdering( O )  then
                ipr := ShallowCopy( IndeterminatesOfPolynomialRing( R ) );
                mcf := MonomialComparisonFunction( O );
                Sort( ipr, mcf );
                ipr := Reversed( ipr );
                R := PolynomialRing( LeftActingDomain( R ), ipr );
            fi;

            if not ( HasTermOrdering( R ) and 
                     IsIdenticalObj( TermOrdering( R ), O ) )  then
                SetTermOrdering( R, O );
                SingularSetBaseRing( R );
            fi;

            I := Ideal( R, pols );
            return GroebnerBasis( I, degb );
        end );

    GBASIS := SINGULARGBASIS;
    GetGroebner := SINGULARGBASIS.GroebnerBasis;
else
    GetGroebner := ReducedGroebnerBasis;
fi;

#############################################################################
##
## some elementary functions
##
B2 := function(x) return x*(x-1)/2; end;
B3 := function(x) return x*(x-1)*(x-2)/6; end;

#############################################################################
##
## set up global vars
##
ORDER := MonomialGrevlexOrdering();
PRR := PolynomialReducedRemainder;
LIMIT := 12;

#############################################################################
##
## set up dummies of vars
## 
NAMES := [];
T := NullMat(2,2);
A := List([1..2], x -> []);
B := List([1..2], x -> []);
x := Indeterminate(Rationals, "x");
y := Indeterminate(Rationals, "y");
HallRec := [];

